﻿using Microsoft.Win32.SafeHandles;
using System;
using System.Collections.Generic;
using System.Linq;
using System.Runtime.InteropServices;
using System.Security.AccessControl;

namespace NtApi
{
  [Flags]
  public enum AttributeFlags : uint
  {
    None = 0,
    Inherit = 0x00000002,
    Permanent = 0x00000010,
    Exclusive = 0x00000020,
    CaseInsensitive = 0x00000040,
    OpenIf = 0x00000080,
    OpenLink = 0x00000100,
    KernelHandle = 0x00000200,
    ForceAccessCheck = 0x00000400,
    IgnoreImpersonatedDevicemap = 0x00000800,
    DontReparse = 0x00001000,
  }

  public class IoStatus
  {
    public IntPtr Pointer;
    public IntPtr Information;
  }

  [Flags]
  public enum ShareMode
  {
    None = 0,
    Read = 0x00000001,
    Write = 0x00000002,
    Delete = 0x00000004,
  }

  [Flags]
  public enum FileOpenOptions
  {
    None = 0,
    DirectoryFile = 0x00000001,
    WriteThrough  = 0x00000002,
    SequentialOnly = 0x00000004,
    NoIntermediateBuffering = 0x00000008,
    SynchronousIoAlert = 0x00000010,
    SynchronousIoNonAlert = 0x00000020,
    NonDirectoryFile = 0x00000040,
    CreateTreeConnection = 0x00000080,
    CompleteIfOplocked = 0x00000100,
    NoEaKnowledge = 0x00000200,
    OpenRemoteInstance = 0x00000400,
    RandomAccess = 0x00000800,
    DeleteOnClose = 0x00001000,
    OpenByFileId  = 0x00002000,
    OpenForBackupIntent = 0x00004000,
    NoCompression = 0x00008000,
    OpenRequiringOplock = 0x00010000,
    ReserveOpfilter = 0x00100000,
    OpenReparsePoint = 0x00200000,
    OpenNoRecall = 0x00400000,
    OpenForFreeSpaceQuery = 0x00800000
  }

  [Flags]
  public enum GenericAccessRights : uint
  {
    None = 0,
    GenericRead = 0x80000000,
    GenericWrite = 0x40000000,
    GenericExecute = 0x20000000,
    GenericAll = 0x10000000,
    Delete = 0x00010000,
		ReadControl = 0x00020000,
		WriteDac = 0x00040000,
		WriteOwner = 0x00080000,
		Synchronize = 0x00100000,
    MaximumAllowed = 0x02000000,
  };

  [Flags]
  public enum ProcessAccessRights : uint
  {
    None = 0,
    CreateProcess = 0x0080,
		CreateThread = 0x0002,
		DupHandle = 0x0040,
		QueryInformation = 0x0400,
		QueryLimitedInformation = 0x1000,
		SetInformation = 0x0200,
		SetQuota = 0x0100,
		SuspendResume = 0x0800,
		Terminate = 0x0001,
		VmOperation = 0x0008,
		VmRead = 0x0010,
		VmWrite = 0x0020,
    MaximumAllowed = GenericAccessRights.MaximumAllowed
	};

  [Flags]
  public enum FileAccessRights : uint
  {
    None = 0,
    ReadData = 0x0001,
		WriteData = 0x0002,
		AppendData = 0x0004,
		ReadEa = 0x0008,
		WriteEa = 0x0010,
		Execute = 0x0020,
		DeleteChild = 0x0040,
		ReadAttributes = 0x0080,
		WriteAttributes = 0x0100,
    GenericRead = 0x80000000,
    GenericWrite = 0x40000000,
    GenericExecute = 0x20000000,
    GenericAll = 0x10000000,
    Delete = 0x00010000,
		ReadControl = 0x00020000,
		WriteDac = 0x00040000,
		WriteOwner = 0x00080000,
		Synchronize = 0x00100000,
    MaximumAllowed = 0x02000000,
	}

  [StructLayout(LayoutKind.Sequential, CharSet=CharSet.Unicode)]
  public sealed class UnicodeString
  {
    ushort Length;
    ushort MaximumLength;
    [MarshalAs(UnmanagedType.LPWStr)] string Buffer;
    
    public UnicodeString(string str)
    {
      Length = (ushort)(str.Length * 2);
      MaximumLength = (ushort)((str.Length * 2) + 1);
      Buffer = str;
    }
  }

  // This class is used when the UNICODE_STRING is an output parameter
  public struct UnicodeStringOut
  {
    public ushort Length;
    public ushort MaximumLength;
    public IntPtr Buffer;
    public override string ToString()
    {
      return Marshal.PtrToStringUni(Buffer, Length);
    }
  }

  public enum SecurityImpersonationLevel
  {
    Anonymous = 0,
    Identification = 1,
    Impersonation = 2,
    Delegation = 3
  }

  public enum SecurityContextTrackingMode : byte
  {
    Static = 0,
    Dynamic = 1
  }

  [StructLayout(LayoutKind.Sequential)]
  public sealed class SecurityQualityOfService
  {
    int Length;
    public SecurityImpersonationLevel ImpersonationLevel;
    public SecurityContextTrackingMode ContextTrackingMode;
    [MarshalAs(UnmanagedType.U1)] public bool EffectiveOnly;

    public SecurityQualityOfService()
    {
      Length = Marshal.SizeOf(this);
    }
  }

  [StructLayout(LayoutKind.Sequential, CharSet = CharSet.Unicode)]
  public sealed class ObjectAttributes : IDisposable
  {
    int Length;
    IntPtr RootDirectory;
    IntPtr ObjectName;
    AttributeFlags Attributes;
    IntPtr SecurityDescriptor;
    IntPtr SecurityQualityOfService;

    private static IntPtr AllocStruct(object s)
    {
      int size = Marshal.SizeOf(s);
      IntPtr ret = Marshal.AllocHGlobal(size);
      Marshal.StructureToPtr(s, ret, false);
      return ret;
    }

    private static void FreeStruct(ref IntPtr p, Type struct_type)
    {
      Marshal.DestroyStructure(p, struct_type);
      Marshal.FreeHGlobal(p);
      p = IntPtr.Zero;
    }

    public ObjectAttributes() : this(AttributeFlags.None)
    {
    }

    public ObjectAttributes(string object_name, AttributeFlags attributes) : this(object_name, attributes, IntPtr.Zero, null, null)
    {
    }

    public ObjectAttributes(AttributeFlags attributes) : this(null, attributes, IntPtr.Zero, null, null)
    {
    }

    public ObjectAttributes(string object_name) : this(object_name, AttributeFlags.CaseInsensitive, IntPtr.Zero, null, null)
    {
    }

    public ObjectAttributes(string object_name, AttributeFlags attributes, IntPtr root, SecurityQualityOfService sqos, GenericSecurityDescriptor security_descriptor)
    {
      Length = Marshal.SizeOf(this);
      if (object_name != null)
      {
        ObjectName = AllocStruct(new UnicodeString(object_name));
      }
      Attributes = attributes;
      if (sqos != null)
      {
        SecurityQualityOfService = AllocStruct(sqos);
      }
      RootDirectory = root;
      if (security_descriptor != null)
      {
        byte[] sd_binary = new byte[security_descriptor.BinaryLength];
        security_descriptor.GetBinaryForm(sd_binary, 0);
        SecurityDescriptor = Marshal.AllocHGlobal(sd_binary.Length);
        Marshal.Copy(sd_binary, 0, SecurityDescriptor, sd_binary.Length);
      }
    }

    public void Dispose()
    {
      if (ObjectName != IntPtr.Zero)
      {
        FreeStruct(ref ObjectName, typeof(UnicodeString));
      }
      if (SecurityQualityOfService != IntPtr.Zero)
      {
        FreeStruct(ref SecurityQualityOfService, typeof(SecurityQualityOfService));
      }
      if (SecurityDescriptor != IntPtr.Zero)
      {
        Marshal.FreeHGlobal(SecurityDescriptor);
        SecurityDescriptor = IntPtr.Zero;
      }
      GC.SuppressFinalize(this);
    }

    ~ObjectAttributes()
    {
      Dispose();
    }
  }

  public enum ProcessCreateState
  {
    InitialState,
    FailOnFileOpen,
    FailOnSectionCreate,
    FailExeFormat,
    FailMachineMismatch,
    FailExeName,
    Success,
  };

  public enum ProcessCreateStateSuccessOutputFlags : uint
  {
    ProtectedProcess = 1,
    AddressSpaceOverride = 2,
    DevOverrideEnabled = 4,
    ManifestDetected = 8,
    ProtectedProcessLight = 16,
  }

  [StructLayout(LayoutKind.Sequential)]
  public struct ProcessCreateStateSuccessData
  {
    public ProcessCreateStateSuccessOutputFlags OutputFlags;
    public IntPtr FileHandle;
    public IntPtr SectionHandle;
    public ulong UserProcessParametersNative;
    public uint UserProcessParametersWow64;
    public uint CurrentParameterFlags;
    public ulong PebAddressNative;
    public uint PebAddressWow64;
    public ulong ManifestAddress;
    public uint ManifestSize;
  }

  [Flags]
  public enum ProcessCreateInitFlag : ushort
  {
    None = 0,
    WriteOutputOnExit = 1,
    DetectManifest = 2,
    IFEOSkipDebugger = 4,
    IFEODoNotPropagateKeyState = 8,
  }

  [StructLayout(LayoutKind.Explicit)]
  public struct ProcessCreateInfoData
  {
    // InitialState
    [FieldOffset(0)]
    public ProcessCreateInitFlag InitFlags;
    [FieldOffset(2)]
    public ImageCharacteristics ProhibitedImageCharacteristics;
    [FieldOffset(4)]
    public FileAccessRights AdditionalFileAccess;
    // FailOnSectionCreate
    [FieldOffset(0)]
    public IntPtr FileHandle;
    // FailExeFormat
    [FieldOffset(0)]
    public ushort DllCharacteristics;
    // FailExeName
    [FieldOffset(0)]
    public IntPtr IFEOKey;
    // Success
    [FieldOffset(0)]
    public ProcessCreateStateSuccessData Success;
  }

  [StructLayout(LayoutKind.Sequential)]
  public class ProcessCreateInfo
  {
    IntPtr Size;
    public ProcessCreateState State;
    public ProcessCreateInfoData Data;

    public ProcessCreateInfo()
    {
      Size = new IntPtr(Marshal.SizeOf(this));
      State = ProcessCreateState.InitialState;
    }
  }

  [StructLayout(LayoutKind.Sequential)]
  public struct RtlDriveLetterCurDir
  {
    ushort Flags;
    ushort Length;
    uint TimeStamp;
    UnicodeStringOut DosPath;
  }

  [StructLayout(LayoutKind.Sequential)]
  public class RtlUserProcessParameters
  {
    public uint MaximumLength;
    public uint Length;
    public uint Flags;
    public uint DebugFlags;
    public IntPtr ConsoleHandle;
    public uint ConsoleFlags;
    public IntPtr StdInputHandle;
    public IntPtr StdOutputHandle;
    public IntPtr StdErrorHandle;
    public UnicodeStringOut CurrentDirectoryPath;
    public IntPtr CurrentDirectoryHandle;
    public UnicodeStringOut DllPath;
    public UnicodeStringOut ImagePathName;
    public UnicodeStringOut CommandLine;
    public IntPtr Environment;
    public uint StartingPositionLeft;
    public uint StartingPositionTop;
    public uint Width;
    public uint Height;
    public uint CharWidth;
    public uint CharHeight;
    public uint ConsoleTextAttributes;
    public uint WindowFlags;
    public uint ShowWindowFlags;
    public UnicodeStringOut WindowTitle;
    public UnicodeStringOut DesktopName;
    public UnicodeStringOut ShellInfo;
    public UnicodeStringOut RuntimeData;
    [MarshalAs(UnmanagedType.ByValArray, SizeConst = 0x20)]
    public RtlDriveLetterCurDir[] DLCurrentDirectory;
  }

  [Flags]
  public enum SectionImageFlags : byte
  {
    ComPlusNativeReady = 1,
    ComPlusILOnly = 2,
    ImageDynamicallyRelocated = 4,
    ImageMappedFlat = 8,
    BaseBelow4gb = 16
  }

  [StructLayout(LayoutKind.Sequential)]
  public class SectionImageInformation
  {
    public IntPtr TransferAddress;
    public uint ZeroBits;
    public IntPtr MaximumStackSize;
    public IntPtr CommittedStackSize;
    public uint SubSystemType;
    public ushort SubSystemMinorVersion;
    public ushort SubSystemMajorVersion;
    public uint GpValue;
    public ushort ImageCharacteristics;
    public ushort DllCharacteristics;
    public ushort Machine;
    [MarshalAs(UnmanagedType.U1)]
    public bool ImageContainsCode;
    [MarshalAs(UnmanagedType.U1)]
    public SectionImageFlags ImageFlags;
    public uint LoaderFlags;
    public uint ImageFileSize;
    public uint CheckSum;
  };

  public enum PsProtectedType
  {
    PsProtectedTypeNone,
    PsProtectedTypeProtectedLight,
    PsProtectedTypeProtected,
    PsProtectedTypeMax
  }

  public enum PsProtectedSigner
  {
    PsProtectedSignerNone,
    PsProtectedSignerAuthenticode,
    PsProtectedSignerCodeGen,
    PsProtectedSignerAntimalware,
    PsProtectedSignerLsa,
    PsProtectedSignerWindows,
    PsProtectedSignerWinTcb,
    PsProtectedSignerMax
  }

  [StructLayout(LayoutKind.Sequential)]
  public class PsProtection
  {
    private byte level;

    public PsProtection(PsProtectedType type, PsProtectedSigner signer, bool audit)
    {
      level = (byte)((int)type | (audit ? 0x8 : 0) | ((int)signer << 4));
    }
  }
  

  [StructLayout(LayoutKind.Sequential)]
  public struct ProcessAttributeNative
  {
    public uint Attribute;
    public IntPtr Size;
    public IntPtr ValuePtr;
    public IntPtr ReturnLength;

    public ProcessAttributeNative(uint attribute, IntPtr valueptr, IntPtr size, IntPtr return_length)
    {
      Attribute = attribute;
      ValuePtr = valueptr;
      Size = size;
      ReturnLength = return_length;
    }
  }

  public enum ProcessAttributeNum
  {
    ParentProcess,
    DebugPort,
    Token,
    ClientId, 
    TebAddress,
    ImageName, 
    ImageInfo, 
    MemoryReserve,
    PriorityClass, 
    ErrorMode,
    StdHandleInfo, 
    HandleList,
    GroupAffinity,
    PreferredNode,
    IdealProcessor,
    UmsThread,
    MitigationOptions,
    ProtectionLevel,
    ChildProcess,
    JobList,
  }

  public class ProcessAttribute : IDisposable
  {
    const uint NumberMask = 0x0000FFFF;
    const uint ThreadOnly = 0x00010000;
    const uint InputOnly = 0x00020000; 
    const uint Additive = 0x00040000;

    SafeHandle _handle;
    ProcessAttributeNum _attribute_num;
    bool _thread;
    bool _input;
    bool _additive;
    IntPtr _valueptr;
    IntPtr _size;
    IntPtr _return_length;

    private ProcessAttribute(ProcessAttributeNum num, bool thread, bool input, bool additive, IntPtr valueptr, int size, IntPtr return_length)
    {
      _attribute_num = num;
      _thread = thread;
      _input = input;
      _additive = additive;
      _valueptr = valueptr;
      _size = new IntPtr(size);
      _return_length = return_length;
    }

    private ProcessAttribute(ProcessAttributeNum num, bool thread, bool input, bool additive, SafeHandle handle, int size, IntPtr return_length) :
       this(num, thread, input, additive, handle.DangerousGetHandle(), size, return_length)
    {
      _handle = handle;
    }

    private ProcessAttribute(ProcessAttributeNum num, bool thread, bool input, bool additive, SafeHGlobalBuffer handle) :
      this(num, thread, input, additive, handle, handle.Length, IntPtr.Zero)
    {
    }

    private ProcessAttribute(ProcessAttributeNum num, bool thread, bool input, bool additive, SafeKernelObjectHandle handle) :
        this(num, thread, input, additive, handle, IntPtr.Size, IntPtr.Zero)
    {
    }

    public ProcessAttributeNative GetNativeAttribute()
    {
      IntPtr valueptr = _handle != null ? _handle.DangerousGetHandle() : _valueptr;
      return new ProcessAttributeNative(((uint)_attribute_num & NumberMask)
        | (_thread ? ThreadOnly : 0) | (_input ? InputOnly : 0) | (_additive ? Additive : 0), valueptr, _size, _return_length);
    }
    
    public static ProcessAttribute ImageName(string image_name)
    {
      SafeHGlobalBuffer name = new SafeHGlobalBuffer(Marshal.StringToHGlobalUni(image_name), image_name.Length * 2, true);
      return new ProcessAttribute(ProcessAttributeNum.ImageName, false, true, false, 
            name);
    }

    public static ProcessAttribute ParentProcess(SafeKernelObjectHandle parent_process)
    {
      return new ProcessAttribute(ProcessAttributeNum.ParentProcess,
        false, true, true, NtSystemCalls.DuplicateObject(parent_process));
    }

    public static ProcessAttribute Token(SafeKernelObjectHandle token)
    {
      return new ProcessAttribute(ProcessAttributeNum.Token,
        false, true, true, NtSystemCalls.DuplicateObject(token));
    }

    public static ProcessAttribute ImageInfo(SafeStructureOutBuffer<SectionImageInformation> image_information)
    {
      return new ProcessAttribute(ProcessAttributeNum.ImageInfo, false, false, false, image_information);
    }

    public static ProcessAttribute ClientId(SafeStructureOutBuffer<ClientId> client_id)
    {
      return new ProcessAttribute(ProcessAttributeNum.ClientId, true, false, false, client_id);
    }

    public static ProcessAttribute ChildProcess(bool child_process_restricted, bool child_process_override)
    {
      int value = child_process_restricted ? 1 : 0;
      if (child_process_override)
      {
        value |= 2;
      }
      return new ProcessAttribute(ProcessAttributeNum.ChildProcess, false, true, false, new IntPtr(value), IntPtr.Size, IntPtr.Zero);
    }

    public static ProcessAttribute ProtectionLevel(PsProtectedType type, PsProtectedSigner signer, bool audit)
    {
      return new ProcessAttribute(ProcessAttributeNum.ProtectionLevel, false, true, true, new SafeStructureBuffer(new PsProtection(type, signer, audit)));
    }

    public static ProcessAttribute HandleList(IEnumerable<SafeHandle> handles)
    {
      return new ProcessAttribute(ProcessAttributeNum.HandleList, false, true, false, 
        new SafeHandleListHandle(handles.Select(h => NtSystemCalls.DuplicateObject(h)), true));
    }

    #region IDisposable Support
    private bool disposedValue = false; // To detect redundant calls

    protected virtual void Dispose(bool disposing)
    {
      if (!disposedValue)
      {
        if (_handle != null)
        {
          _handle.Close();
          _handle = null;
        }

        disposedValue = true;
      }
    }

    ~ProcessAttribute()
    {
      Dispose(false);
    }

    public void Dispose()
    {
      Dispose(true);
      GC.SuppressFinalize(this);
    }
    #endregion

  };

  [StructLayout(LayoutKind.Sequential)]
  public sealed class ProcessAttributeList 
  {
    IntPtr TotalLength;
    // Allocate upto 64 entries.
    [MarshalAs(UnmanagedType.ByValArray, SizeConst = 64)]
    ProcessAttributeNative[] Attributes;

    public ProcessAttributeList(IEnumerable<ProcessAttribute> attributes)
    {
      ProcessAttributeNative[] attrs = attributes.Select(a => a.GetNativeAttribute()).ToArray();
      Attributes = new ProcessAttributeNative[64];
      Array.Copy(attrs, Attributes, attrs.Length);      
      TotalLength = new IntPtr(IntPtr.Size + Marshal.SizeOf(typeof(ProcessAttributeNative)) * attrs.Length);
    }
  }

  [Flags]
  public enum ProcessCreateFlags
  {
    None = 0,
    BreakawayJob = 0x00000001,    // Only allowed if job allows breakaway
    NoDebugInherit = 0x00000002,
    InheritHandles = 0x00000004,
    OverrideAddressSpace = 0x00000008,
    LargePages = 0x00000010,
    LargePagesSystemDll = 0x00000020,
    ProtectedProcess = 0x00000040,
    CreateSession = 0x00000080,
    InheritFromParent = 0x00000100,
    TerminateOnAbnormalExit = 0x200,
    ForceBreakawayJob = 0x400,     // Needs TCB
    // This is same as ForceBreakawayJob prior to Anniversary Edition
    MinimalProcess = 0x800,        // Needs Kernel Privileges
    IgnoreSectionObject = 0x1000,
    MinimalProcessFlag1 = 0x2000,
    MinimalProcessFlag2 = 0x4000,
    AuxiliaryProcess = 0x8000,     // Needs TCB
  }

  [Flags]
  public enum ThreadCreateFlags
  {
    None = 0,
    Suspended = 0x00000001,
    SkipThreadAttach = 0x00000002,
    HideFromDebugger = 0x00000004,
    HasSecurityDescriptor = 0x00000010,
    AccessCheckInTarget = 0x00000020,
    InitialThread = 0x00000080,
  }

  [StructLayout(LayoutKind.Sequential)]
  public class ProcessBasicInformation
  {
    public int ExitStatus;
    public IntPtr PebBaseAddress;
    public IntPtr AffinityMask;
    public int BasePriority;
    public IntPtr UniqueProcessId;
    public IntPtr InheritedFromUniqueProcessId;
  }

  [StructLayout(LayoutKind.Sequential)]
  public class ProcessSessionInformation
  {
    public int SessionId;
  }

  public enum ProcessInfoClass
  {
    ProcessBasicInformation, 
    ProcessQuotaLimits, 
    ProcessIoCounters, 
    ProcessVmCounters, 
    ProcessTimes, 
    ProcessBasePriority, 
    ProcessRaisePriority, 
    ProcessDebugPort, 
    ProcessExceptionPort, 
    ProcessAccessToken, 
    ProcessLdtInformation, 
    ProcessLdtSize, 
    ProcessDefaultHardErrorMode, 
    ProcessIoPortHandlers, 
    ProcessPooledUsageAndLimits, 
    ProcessWorkingSetWatch, 
    ProcessUserModeIOPL,
    ProcessEnableAlignmentFaultFixup, 
    ProcessPriorityClass, 
    ProcessWx86Information,
    ProcessHandleCount, 
    ProcessAffinityMask, 
    ProcessPriorityBoost, 
    ProcessDeviceMap, 
    ProcessSessionInformation, 
    ProcessForegroundInformation, 
    ProcessWow64Information, 
    ProcessImageFileName, 
    ProcessLUIDDeviceMapsEnabled, 
    ProcessBreakOnTermination, 
    ProcessDebugObjectHandle, 
    ProcessDebugFlags, 
    ProcessHandleTracing, 
    ProcessIoPriority, 
    ProcessExecuteFlags, 
    ProcessResourceManagement,
    ProcessCookie, 
    ProcessImageInformation, 
    ProcessCycleTime, 
    ProcessPagePriority, 
    ProcessInstrumentationCallback, 
    ProcessThreadStackAllocation, 
    ProcessWorkingSetWatchEx, 
    ProcessImageFileNameWin32, 
    ProcessImageFileMapping, 
    ProcessAffinityUpdateMode, 
    ProcessMemoryAllocationMode, 
    ProcessGroupInformation, 
    ProcessTokenVirtualizationEnabled, 
    ProcessConsoleHostProcess, 
    ProcessWindowInformation, 
    ProcessHandleInformation, 
    ProcessMitigationPolicy, 
    ProcessDynamicFunctionTableInformation,
    ProcessHandleCheckingMode,
    ProcessKeepAliveCount, 
    ProcessRevokeFileHandles, 
    ProcessWorkingSetControl, 
    ProcessHandleTable, 
    ProcessCheckStackExtentsMode,
    ProcessCommandLineInformation, 
    ProcessProtectionInformation, 
    ProcessMemoryExhaustion, 
    ProcessFaultInformation, 
    ProcessTelemetryIdInformation, 
    ProcessCommitReleaseInformation, 
    ProcessDefaultCpuSetsInformation,
    ProcessAllowedCpuSetsInformation,
    ProcessReserved1Information,
    ProcessReserved2Information,
    ProcessSubsystemProcess, 
    ProcessJobMemoryInformation, 
  }

  public sealed class SafeKernelObjectHandle
    : SafeHandleZeroOrMinusOneIsInvalid
  {
    public SafeKernelObjectHandle() 
      : base(true)
    {
    }

    public SafeKernelObjectHandle(IntPtr handle, bool owns_handle)
      : base(owns_handle)
    {
      SetHandle(handle);
    }

    protected override bool ReleaseHandle()
    {
      if (!IsInvalid)
      {
        NtSystemCalls.NtClose(this.handle);
        this.handle = IntPtr.Zero;
        return true;
      }
      return false;
    }
  }

  public sealed class SafeHandleListHandle : SafeHGlobalBuffer
  {
    private SafeKernelObjectHandle[] _handles;

    public SafeHandleListHandle(IEnumerable<SafeKernelObjectHandle> handles, bool owns_handles) 
      : base(IntPtr.Size * handles.Count())
    {
      _handles = handles.ToArray();
      IntPtr buffer = handle;
      for (int i = 0; i < _handles.Length; ++i)
      {
        Marshal.WriteIntPtr(buffer, _handles[i].DangerousGetHandle());
        buffer += IntPtr.Size;
      }
    }

    protected override bool ReleaseHandle()
    {
      foreach (SafeKernelObjectHandle handle in _handles)
      {
        handle.Close();
      }
      _handles = new SafeKernelObjectHandle[0];
      return base.ReleaseHandle();
    }
  }


  [StructLayout(LayoutKind.Explicit)]
  public class LargeInteger
  {
    [FieldOffset(0)]
    public uint LowPart;
    [FieldOffset(4)]
    public int HighPart;
    [FieldOffset(0)]
    public long QuadPart;
  }

  [Flags]
  public enum SectionAccessRights : uint
  {
    Query = 0x0001,
    MapWrite = 0x0002,
    MapRead = 0x0004,
    MapExecute = 0x0008,
    ExtendSize = 0x0010,
    MapExecuteExplicit = 0x0020
  }

  public enum ProtectionType
  {
    None = 0,
    NoAccess = 0x01,
    ReadOnly = 0x02,
    ReadWrite = 0x04,
    WriteCopy = 0x08,
    Execute = 0x10,
    ExecuteRead = 0x20,
    ExecuteWriteCopy = 0x80
  }

  [Flags]
  public enum SectionAttributes
  {
    None = 0,
    Based = 0x00200000,
    NoChange = 0x00400000,
    Image = 0x01000000,
    Reserver = 0x04000000,
    Commit = 0x08000000,
    NoCache = 0x10000000,
  }

  public class NtException : ExternalException
  {
    [DllImport("kernel32.dll", CharSet=CharSet.Unicode, SetLastError = true)]
    private static extern IntPtr GetModuleHandle(string modulename);

    [Flags]
    enum FormatFlags
    {
      AllocateBuffer = 0x00000100,
      FromHModule = 0x00000800,
      FromSystem = 0x00001000,
      IgnoreInserts = 0x00000200
    }

    [DllImport("kernel32.dll", CharSet=CharSet.Unicode, SetLastError= true)]
    private static extern int FormatMessage(
      FormatFlags   dwFlags,
      IntPtr lpSource,
      int   dwMessageId,
      int   dwLanguageId,
      out IntPtr  lpBuffer,
      int   nSize,
      IntPtr  Arguments
    );

    [DllImport("kernel32.dll")]
    private static extern IntPtr LocalFree(IntPtr p);

    private static string StatusToString(int status)
    {
      IntPtr buffer = IntPtr.Zero;
      try
      {
        if (FormatMessage(FormatFlags.AllocateBuffer | FormatFlags.FromHModule | FormatFlags.FromSystem | FormatFlags.IgnoreInserts,
            GetModuleHandle("ntdll.dll"), status, 0, out buffer, 0, IntPtr.Zero) > 0)
        {
          return Marshal.PtrToStringUni(buffer);
        }
      }
      finally
      {
        if (buffer != IntPtr.Zero)
        {
          LocalFree(buffer);
        }
      }
      return String.Format("Unknown Error: 0x{0:X08}", status);
    }

    public NtException(int status) : base(StatusToString(status))
    {
    }
  }

  public class SafeHGlobalBuffer : SafeHandleZeroOrMinusOneIsInvalid
  {
    public SafeHGlobalBuffer(int length) 
      : this(Marshal.AllocHGlobal(length), length, true)
    {
    }

    public SafeHGlobalBuffer(IntPtr buffer, int length, bool owns_handle) 
      : base(owns_handle)
    {
      Length = length;
      SetHandle(buffer);
    }

    public int Length
    {
      get; private set;
    }

    protected override bool ReleaseHandle()
    {
      if (!IsInvalid)
      {
        Marshal.FreeHGlobal(handle);
        handle = IntPtr.Zero;
      }
      return true;
    }
  }

  public class SafeStructureBuffer : SafeHGlobalBuffer
  {
    Type _type;

    public SafeStructureBuffer(object value) : base(Marshal.SizeOf(value))
    {
      _type = value.GetType();
      Marshal.StructureToPtr(value, handle, false);
    }

    protected override bool ReleaseHandle()
    {
      if (!IsInvalid)
      {
        Marshal.DestroyStructure(handle, _type);
      }
      return base.ReleaseHandle();
    }
  }

  public class SafeStructureOutBuffer<T> : SafeHGlobalBuffer
  {
    public SafeStructureOutBuffer() : base(Marshal.SizeOf(typeof(T)))
    {
    }

    public T Result
    {
      get
      {
        if (IsInvalid)
          throw new ObjectDisposedException("handle");

        return Marshal.PtrToStructure<T>(handle);
      }
    }
  }

  [StructLayout(LayoutKind.Sequential)]
  struct FileDispositionInformation
  {
    [MarshalAs(UnmanagedType.U1)]
    public bool DeleteFile;
  }

  [StructLayout(LayoutKind.Sequential, CharSet = CharSet.Unicode)]
  public class FileLinkRenameInformation
  {
    [MarshalAs(UnmanagedType.U1)]
    public bool ReplaceIfExists;
    public IntPtr RootDirectory;
    public int FileNameLength;
    [MarshalAs(UnmanagedType.ByValTStr, SizeConst = 256)]
    public string FileName;
  }

  [StructLayout(LayoutKind.Sequential)]
  public class ClientId
  {
    public IntPtr UniqueProcess;
    public IntPtr UniqueThread;
  }

  public enum TokenType
  {
    Primary = 1,
    Impersonation = 2
  }

  [Flags]
  public enum DuplicateObjectOptions
  {
    CloseSource = 1,
    SameAccess = 2,
    SameAttributes = 4,
  }

  [Flags]
  public enum ImageCharacteristics : ushort
  {
    None = 0,
    RelocsStripped = 0x0001,
    ExecutableImage = 0x0002,
    LineNumsStripped = 0x0004,
    LocalSymsStripped = 0x0008,
    AggresiveWsTrim = 0x0010,
    LargeAddressAware = 0x0020,
    FileBytesReservedLo = 0x0080,
    Image32BitMachine = 0x0100,
    DebugStripped = 0x0200,
    RemovableRunFromSwap = 0x0400,
    NetRunFromSwap = 0x0800,
    System = 0x1000,
    Dll = 0x2000,
    UpSystemOnly = 0x4000,
    BytesReservedHi = 0x8000,
  }


  public enum SystemInformationClass
  {
    SystemBasicInformation, 
    SystemProcessorInformation, 
    SystemPerformanceInformation, 
    SystemTimeOfDayInformation, 
    SystemPathInformation, 
    SystemProcessInformation, 
    SystemCallCountInformation, 
    SystemDeviceInformation, 
    SystemProcessorPerformanceInformation, 
    SystemFlagsInformation, 
    SystemCallTimeInformation, 
    SystemModuleInformation, 
    SystemLocksInformation,
    SystemStackTraceInformation,
    SystemPagedPoolInformation, 
    SystemNonPagedPoolInformation, 
    SystemHandleInformation, 
    SystemObjectInformation, 
    SystemPageFileInformation, 
    SystemVdmInstemulInformation, 
    SystemVdmBopInformation, 
    SystemFileCacheInformation, 
    SystemPoolTagInformation, 
    SystemInterruptInformation, 
    SystemDpcBehaviorInformation, 
    SystemFullMemoryInformation, 
    SystemLoadGdiDriverInformation, 
    SystemUnloadGdiDriverInformation, 
    SystemTimeAdjustmentInformation, 
    SystemSummaryMemoryInformation, 
    SystemMirrorMemoryInformation, 
    SystemPerformanceTraceInformation, 
    SystemObsolete0, 
    SystemExceptionInformation, 
    SystemCrashDumpStateInformation, 
    SystemKernelDebuggerInformation, 
    SystemContextSwitchInformation, 
    SystemRegistryQuotaInformation, 
    SystemExtendServiceTableInformation, 
    SystemPrioritySeperation, 
    SystemVerifierAddDriverInformation, 
    SystemVerifierRemoveDriverInformation, 
    SystemProcessorIdleInformation, 
    SystemLegacyDriverInformation, 
    SystemCurrentTimeZoneInformation, 
    SystemLookasideInformation, 
    SystemTimeSlipNotification, 
    SystemSessionCreate, 
    SystemSessionDetach, 
    SystemSessionInformation, 
    SystemRangeStartInformation, 
    SystemVerifierInformation, 
    SystemVerifierThunkExtend, 
    SystemSessionProcessInformation, 
    SystemLoadGdiDriverInSystemSpace, 
    SystemNumaProcessorMap, 
    SystemPrefetcherInformation, 
    SystemExtendedProcessInformation, 
    SystemRecommendedSharedDataAlignment, 
    SystemComPlusPackage, 
    SystemNumaAvailableMemory, 
    SystemProcessorPowerInformation, 
    SystemEmulationBasicInformation, 
    SystemEmulationProcessorInformation,
    SystemExtendedHandleInformation, 
    SystemLostDelayedWriteInformation, 
    SystemBigPoolInformation, 
    SystemSessionPoolTagInformation, 
    SystemSessionMappedViewInformation, 
    SystemHotpatchInformation, 
    SystemObjectSecurityMode, 
    SystemWatchdogTimerHandler, 
    SystemWatchdogTimerInformation, 
    SystemLogicalProcessorInformation, 
    SystemWow64SharedInformationObsolete, 
    SystemRegisterFirmwareTableInformationHandler, 
    SystemFirmwareTableInformation, 
    SystemModuleInformationEx, 
    SystemVerifierTriageInformation, 
    SystemSuperfetchInformation, 
    SystemMemoryListInformation, 
    SystemFileCacheInformationEx, 
    SystemThreadPriorityClientIdInformation, 
    SystemProcessorIdleCycleTimeInformation, 
    SystemVerifierCancellationInformation, 
    SystemProcessorPowerInformationEx, 
    SystemRefTraceInformation, 
    SystemSpecialPoolInformation, 
    SystemProcessIdInformation, 
    SystemErrorPortInformation, 
    SystemBootEnvironmentInformation, 
    SystemHypervisorInformation, 
    SystemVerifierInformationEx, 
    SystemTimeZoneInformation, 
    SystemImageFileExecutionOptionsInformation, 
    SystemCoverageInformation, 
    SystemPrefetchPatchInformation, 
    SystemVerifierFaultsInformation, 
    SystemSystemPartitionInformation, 
    SystemSystemDiskInformation, 
    SystemProcessorPerformanceDistribution, 
    SystemNumaProximityNodeInformation, 
    SystemDynamicTimeZoneInformation, 
    SystemCodeIntegrityInformation, 
    SystemProcessorMicrocodeUpdateInformation, 
    SystemProcessorBrandString, 
    SystemVirtualAddressInformation, 
    SystemLogicalProcessorAndGroupInformation, 
    SystemProcessorCycleTimeInformation, 
    SystemStoreInformation, 
    SystemRegistryAppendString, 
    SystemAitSamplingValue, 
    SystemVhdBootInformation, 
    SystemCpuQuotaInformation, 
    SystemNativeBasicInformation, 
    SystemSpare1, 
    SystemLowPriorityIoInformation, 
    SystemTpmBootEntropyInformation, 
    SystemVerifierCountersInformation, 
    SystemPagedPoolInformationEx, 
    SystemSystemPtesInformationEx, 
    SystemNodeDistanceInformation, 
    SystemAcpiAuditInformation, 
    SystemBasicPerformanceInformation, 
    SystemQueryPerformanceCounterInformation, 
    SystemSessionBigPoolInformation, 
    SystemBootGraphicsInformation,
    SystemScrubPhysicalMemoryInformation,
    SystemBadPageInformation,
    SystemProcessorProfileControlArea,
    SystemCombinePhysicalMemoryInformation, 
    SystemEntropyInterruptTimingCallback,
    SystemConsoleInformation,
    SystemPlatformBinaryInformation,
    SystemThrottleNotificationInformation,
    SystemHypervisorProcessorCountInformation,
    SystemDeviceDataInformation,
    SystemDeviceDataEnumerationInformation,
    SystemMemoryTopologyInformation,
    SystemMemoryChannelInformation,
    SystemBootLogoInformation, 
    SystemProcessorPerformanceInformationEx, 
    SystemSpare0,
    SystemSecureBootPolicyInformation,
    SystemPageFileInformationEx, 
    SystemSecureBootInformation,
    SystemEntropyInterruptTimingRawInformation,
    SystemPortableWorkspaceEfiLauncherInformation,
    SystemFullProcessInformation, 
    SystemKernelDebuggerInformationEx, 
    SystemBootMetadataInformation, 
    SystemSoftRebootInformation,
    SystemElamCertificateInformation,
    SystemOfflineDumpConfigInformation,
    SystemProcessorFeaturesInformation, 
    SystemRegistryReconciliationInformation,
    SystemEdidInformation,
    SystemManufacturingInformation, 
    SystemEnergyEstimationConfigInformation, 
    SystemHypervisorDetailInformation, 
    SystemProcessorCycleStatsInformation, 
    SystemVmGenerationCountInformation,
    SystemTrustedPlatformModuleInformation, 
    SystemKernelDebuggerFlags,
    SystemCodeIntegrityPolicyInformation,
    SystemIsolatedUserModeInformation,
    SystemHardwareSecurityTestInterfaceResultsInformation,
    SystemSingleModuleInformation, 
    SystemAllowedCpuSetsInformation,
    SystemDmaProtectionInformation,
    SystemInterruptCpuSetsInformation,
    SystemSecureBootPolicyFullInformation,
    SystemCodeIntegrityPolicyFullInformation,
    SystemAffinitizedInterruptProcessorInformation,
    SystemRootSiloInformation, 
    MaxSystemInfoClass
  }

  [StructLayout(LayoutKind.Sequential)]
  public class SystemHandleTableInfoEntry
  {
    public ushort UniqueProcessId;
    public ushort CreatorBackTraceIndex;
    public byte ObjectTypeIndex;
    public byte HandleAttributes;
    public ushort HandleValue;
    public IntPtr Object;
    public uint GrantedAccess;
  }

  public enum TokenInformationClass
  {
    TokenUser = 1,
    TokenGroups = 2,
    TokenPrivileges = 3,
    TokenOwner = 4,
    TokenPrimaryGroup = 5,
    TokenDefaultDacl = 6,
    TokenSource = 7,
    TokenType = 8,
    TokenImpersonationLevel = 9,
    TokenStatistics = 10,
    TokenRestrictedSids = 11,
    TokenSessionId = 12,
    TokenGroupsAndPrivileges = 13,
    TokenSessionReference = 14,
    TokenSandBoxInert = 15,
    TokenAuditPolicy = 16,
    TokenOrigin = 17,
    TokenLinkedToken = 19,
    TokenElevation = 20,
    TokenHasRestrictions = 21,
    TokenAccessInformation = 22,
    TokenVirtualizationAllowed = 23,
    TokenVirtualizationEnabled = 24,
    TokenIntegrityLevel = 25,
    TokenUIAccess = 26,
    TokenMandatoryPolicy = 27,
    TokenLogonSid = 28,
    TokenIsAppContainer = 29,
    TokenCapabilities = 30,
    TokenAppContainerSid = 31,
    TokenAppContainerNumber = 32,
    TokenUserClaimAttributes = 33,
    TokenDeviceClaimAttributes = 34,
    TokenRestrictedUserClaimAttributes = 35,
    TokenRestrictedDeviceClaimAttributes = 36,
    TokenDeviceGroups = 37,
    TokenRestrictedDeviceGroups = 38,
    TokenSecurityAttributes = 39,
    TokenIsRestricted = 40,
    TokenProcessTrustLevel = 41,
    MaxTokenInfoClass = 42
  }
}
